import pandas as pd 
from datetime import datetime 

from core import GPMClient 
from dates import get_missing_dates

def initialize_client(client: GPMClient, plants_filename: str, total_filename: str, inverter_filename: str,
                      tracker_filename: str, concepts=['total, inverter, tracker'], username=None, password=None):
    """
    Initializes GPM client with all relevant Plant and DataSource information. 
    """
    plants = pd.read_csv(plants_filename)
    total_ds = pd.read_csv(total_filename)
    inverter_ds = pd.read_csv(inverter_filename)
    tracker_ds = pd.read_csv(tracker_filename)

    datasources = {}
    all_datasources = [] 
    for plant_id in plants['PlantId'].unique():
        total = total_ds[total_ds['PlantId'] == plant_id]['DataSourceId'].to_list()
        inverter = inverter_ds[inverter_ds['PlantId'] == plant_id]['DataSourceId'].to_list()
        tracker = tracker_ds[tracker_ds['PlantId'] == plant_id]['DataSourceId'].to_list()

        datasources[plant_id] = {
            'total': total,
            'inverter': inverter,
            'tracker': tracker
        }

        all_datasources.extend(total + inverter + tracker)

    plant_ids = plants['PlantId'].to_list()
    gpm = GPMClient(username, password,
                    plant_ids=plant_ids,
                    concepts=concepts,
                    datasources=datasources)
    gpm.all_datasources = all_datasources

    return gpm

def get_data_for_range(client: GPMClient, start_date: [str, datetime.datetime], end_date: [str, datetime.datetime],
    datasourceids=None, plant_id=None, concept=None, grouping='minute', granularity=1, existing_df=None, return_df=False):
    """ 
    Simple wrapper function that uses "GPMClient.get_data_list_in_batches" to fetch measurements for a given date range.
    
    """
    # Filter DataSourceIds 
    datasourceids = client.filter_datasourceids(
        datasourceids=datasourceids,
        plant_id=plant_id,
        concept=concept
    )
    
    # Check for data in existing DataFrame if given.
    if existing_df is not None:
        start, end = get_missing_dates(existing_df, start_date, end_date,
            plant_id=plant_id, grouping=grouping, granularity=granularity)
        
        # Existing dataframe is complete
        if start is None and end is None:
            return existing_df
        
        # Else we need to fetch missing data 
        start_date = start 
        end_date = end 
        
    # Get data from API
    data = client.get_data_list_in_batches(
        datasourceids, start_date, end_date,
        grouping=grouping, granularity=granularity)
    
    # Return new expanded dataframe 
    if return_df:
        data = pd.concat([existing_df, pd.DataFrame(data)]) if existing_df is not None else pd.DataFrame(data)
        data = data.drop_duplicates(subset=['Date', 'DataSourceId'])
    
    # Else return raw data
    return data

def get_last_data(client: GPMClient, datasourceids=None, plant_id=None, concept=None, return_df=True):
    """
    Fetches last data using "GPMClient.get_last_data_for_all_plants".
    
    TO DO:
        - adapt GPMClient.get_last_data_for_all_plants to use the new DataSourceIds filtering functions.
    """
    pass


import pandas as pd 
from datetime import datetime
from dateutil.relativedelta import relativedelta

from core import PRMTEClient

def get_coordinados(client: PRMTEClient):
    """ 
    Get all "coordinados" in the PRMTE database. 
    
    A "coordinado" is a company that operates in the Chilean electricity market 
    under the instruction of the Coordinador Electrico Nacional (Chile's ISO).

    Args:
        client: PRMTEClass object.

    Returns: 
        A list of dictionaries containing all idCoordinados in the PRMTE database.
    """
    return client.make_api_call('coordinados')

def get_canales(client: PRMTEClient):
    """ 
    Get all available channels in a standard PRMTE meter.

    Args:
        client: PRMTEClass object.

    Returns:
        A list of dictionaries containing all available channels and their description.
    """
    return client.make_api_call('canales')

def get_puntomedidas(client: PRMTEClient, idCoordinado: str):
    """ 
    Get all measure points associated to a single idCoordinado.

    Args:
        client: PRMTEClass object.
        idCoordinado (str): unique identifier for coordinated company.

    Returns:
        A list of dictionaries, each containing a measure point with the 
        following structure and information:
            {
              'idPuntoMedida': the measure point unique identifier.
              'idCoordinado':  the coordinated company id.
              'subestacion':   name of the transmission/distribution
                               substation associated with the measure point.
              'tension':       voltage of the substation (in kV)
              'region' :       region for the measure point.
            }

    """
    return client.make_api_call('puntomedidas', params={'idCoordinado': idCoordinado})

def get_measurements(client: PRMTEClient, idPuntoMedida: str, period: str, end_period=None, granularity='1H', df_format='consolidated'):
    """ 
    Get all measurements for a single measurement point, given a period and optionally an end period.

    Args:
        client: PRMTEClass object.
        idPuntoMedida (str): the measure point unique identifier.
        periodo (str): measurement period (monthly granularity) in YYYYMM format.
        end_period (str): end measurement period in YYYYMM format (optional).
        granularity (str): resampling rule as of Pandas documentation (see 
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases)
        df_format (str): either 'consolidated' or 'columns'. See 'transform_records' function in 
            data package for more information on this parameter.

    Returns:
        A Pandas DataFrame containing all measurements for a single measurement point. If end_period is not None, 
        the DataFrame contains all periods between 'period' and 'end_period'.
    """
    pass
    

def get_historic_measurements(client: PRMTEClient, idPuntoMedida: str, broken_periods=[], granularity='1H', df_format='consolidated'):
    """
    Get all measurements up until the last measurement available. 

    Args:
        client: PRMTEClass object.
        idPuntoMedida (str): the measure point unique identifier.
        broken_periods (list): some periods are unavailable in the PRMTE database for some reason
            (server error). We deliberately skip these records.
        granularity (str): resampling rule as of Pandas documentation (see 
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases)
        df_format (str): either 'consolidated' or 'columns'. See 'transform_records' function in 
            data package for more information on this parameter.
        

    Returns: 
        A Pandas DataFrame with the historic data for that measurement point.
    """
    i = 0
    now = datetime.now()
    total = pd.DataFrame()

    while True:
        period = (now - relativedelta(months=i)).strftime("%Y%m"); i += 1
        records, last_reading = client.get_15min_readings(idPuntoMedida, period)

        if not records:
            if period in broken_periods: continue 
            else: break

        df = transform_records(records, last_reading, format=df_format).reset_index()
        df = df.drop(columns=['idPuntoMedida']).set_index('date').resample(granularity).sum()
        total = pd.concat([total, df], axis=0)

    return total





import pandas as pd 

field_mapper = {
        'COM STATUS': 'Com Status',
        'COMS STATUS': 'Com Status',
        'Plant Com Status': 'Com Status',
        'Plant Energy': 'Energy',
        'Total Meter Energy': 'Energy',
        'Total Meters Energy': 'Energy',
        'Plant Power': 'Power',
        'Total Meter Power': 'Power',
        'Total Meters Power': 'Power',
        'Plant Irradiance': 'Irradiance'
    }

def merge_and_pivot(df, ds, index_col=['PlantId', 'Date'], merge_col='DataSourceId', field_col='DataSourceName',
                    value_col='Value', duplicate_col=['DataSourceId', 'Date'], field_mapper=None, pivot=True, set_datetime=True):
    """
    Merges two dataframes on a common column and applies a pivot operation, resulting in a dataframe with column
    names that correspond to the values present in the "ds" dataframe for the "field_col" column. Optionally, it 
    modifies and filters these values depending on the "field_mapper" dict, dropping all column names (and 
    therefore all data in the "df" dataframe) that are not present in the dict keys. 

    Args:
        df (Pandas DataFrame): contains data points indexed by PlantId and Date, each data point corresponding to 
            a unique combination of the tuple (Date, DataSourceId).
        ds (Pandas DataFrame): contains DataSource information, like DataSourceId and DataSourceName. 
        index_col (list or str): index columns to set in the pivoted dataframe.
        merge_col (str): common column between the data and information dataframes (DataSourceId by default).
        field_col (str): column to use as column names in the pivoted dataframe.
        value_col (str): column tu use as values in the pivoted dataframe.
        duplicate_col (str): subset of columns to drop duplicated entries.
        field_mapper (dict): dict containing original DataSourceNames as keys and cleaned (homogeneous)
            column names as values. Every data point that das a DataSourceName that is not present in the 
            dictionary gets dropped after the merge operation.
        pivot (bool): to pivot the dataframe or not. If not, returns the merged dataframe.
        set_datetime (bool): to set the "Date" column to DateTime or not in the data dataframe ("df"). 

    Returns:
        A Pandas DataFrame.  
    """
    # Drop duplicate rows
    if duplicate_col:
        df.drop_duplicates(subset=duplicate_col, inplace=True)

    # Set datetime column
    if set_datetime:
        df['Date'] = pd.to_datetime(df['Date'])
        
    # Set merge column as index to speed up merge operation
    df.set_index(merge_col, inplace=True)
    ds.set_index(merge_col, inplace=True)

    # Merge
    merged = pd.merge(df, ds, on=merge_col, how='inner')

    # Clean and filter DataSourceNames if mapper passed 
    if field_mapper:
        merged[field_col] = merged[field_col].map(field_mapper).fillna('drop')
        merged = merged[merged[field_col] != 'drop']

    # Return merged dataframe if chosen
    if not pivot: 
        # Set index column if given
        if index_col:
            return merged.reset_index().set_index(index_col)
        return merged

    # Pivot
    pivoted = merged.pivot_table(index=index_col, columns=field_col, values=value_col)
    pivoted.columns.name = None

    return pivoted

def format_last_data(response, datasourceids, tuples=False):
    """
    Filters a response coming from the 'get_last_data' method in the GPMClient class in core package.
    The method is used because the endpoint 'get_last_data' returns the last available measurement 
    for ALL DataSources associated to a single plant, and we're usually interested in only some of them.

    Args:
        response (list): list of dictionaries containing measurement records.
        datasourceids (list): list of DataSourceIds to keep from the response 

    Returns:
        If 'tuples', returns a list of formatted tuples, else it returns a list of dictionaries.
    """
    # For performance 
    datasourceids = set(datasourceids)

    if not tuples:
        return [
            {
                'DataSourceId': item['DataSourceId'],
                #'DataSourceName': item['DataSourceName'],
                'Date': item['LastValue']['Date'],
                'Value': item['LastValue']['Value']
            }
            for item in response if item['DataSourceId'] in datasourceids
        ]
    
    else:
        return [
            (   
                item['LastValue']['Date'].replace('T', ' ').split('.')[0],
                int(item['DataSourceId']),
                float(item['LastValue']['Value'])
            )
            for item in response if item['DataSourceId'] in datasourceids
        ]




import pandas as pd 

def transform_records(records, format='consolidated'):
    """
    Transforms records coming from the 'get_15min_readings' method from PRMTEClient class in core package. 

    Args:
        records (list): list of tuples containing measurement records. See 'get_15min_readings' method.
        format (str): format of output Pandas DataFrame. Either 'consolidated' or 'columns'.

    Returns: 
        A Pandas DataFrame containing all measurement records. Depending on the format chosen:

            if 'consolidated', output DataFrame has a DateTimeIndex (date), with columns "idPuntoMedida"
            (measurement point unique id) and "value" (net energy flow for that interval, in kWh units.
            The value is negative for net imports and positive for net exports.)

            if 'columns', output DataFrame has columns "idPuntoMedida", date (DateTime column), "Retiros"
            for energy withdrawls in kWh and "Inyecciones" for energy injections during that interval.
    """
    df = pd.DataFrame(records, columns=['idPuntoMedida', 'canalVal', 'date', 'value'])
    df['date'] = pd.to_datetime(df['date'])
    df = df.set_index('date').sort_index().loc[:last_reading].reset_index()
    
    # By default withdrawls are negative values indicating the direction of energy flow
    df['value'] = df.apply(lambda row: -row['value'] if row['canalVal'] == 1 else row['value'], axis=1)

    if format == 'consolidated':
        # Not actually invalid, you can inject and withdraw energy during the same period
        #invalid_groups = df.groupby(['idPuntoMedida', 'date']).filter(lambda group: group['value'].ne(0).sum() > 1)
        df = df.groupby(['idPuntoMedida', 'date']).sum().reset_index().drop(columns=['canalVal'])
        df.set_index('date', inplace=True)

    elif format == 'columns':
        df = df.pivot_table(index=['idPuntoMedida', 'date'], columns='canalVal', values='value', fill_value=0).reset_index()
        df.rename(columns={1: 'Retiros', 3: 'Inyecciones'}, inplace=True)

    return df


import os
import logging
import requests
import numpy as np

class PRMTEClient:
    BASE_URL = "https://mercados.api.coordinador.cl/medidas/api/"
    ENDPOINTS = ['canales', 'coordinados', 'puntomedidas', 'medidas']

    def __init__(self, api_key=None):
        self.api_key = api_key or os.environ.get('PRMTE_API_KEY')
        if not self.api_key:
            raise ValueError("API key must be provided or set as an environment variable 'PRMTE_API_KEY'")

    def make_api_call(self, endpoint, params=None, verbose=False):
        if endpoint not in self.ENDPOINTS:
            raise ValueError("Invalid endpoint")

        # Construct the full URL
        url = self.BASE_URL + endpoint
        if params is None:
            params = {}

        params['user_key'] = self.api_key
        response = requests.get(url, params=params)
        if verbose: print(response.url)

        # Check for successful status code
        if response.status_code == 200:
            return response.json()
        else:
            logging.error(f"{response.status_code}: {response.text}")
            return None

    def get_15min_readings(self, idPuntoMedida, periodo, last_reading=False):
        """
        Calls the "medidas" endpoint and fetches all 15min readings for a given period.
        For now, hardcoded to include only active energy channels (injections and withdrawls).

        Args: 
            idPuntoMedida (str): the measure point unique identifier.
            periodo (str): measurement period (monthly granularity) in YYYYMM format.

        Returns:
            List of records (tuples) with the format (idPuntoMedida, idCanal, date, value).
            idCanal = 1 corresponds to active energy withdrawls, while idCanal = 3 is for 
            active energy injections.
        
        Units:
            Date comes in "YYYY-MM-DD HH:MM:SS" format. 
            Value is in kWh units.
        """
        params = {
            'idPuntoMedida': idPuntoMedida,
            'periodo': periodo,
            'idCanal': '1,3'
        }

        res = self.make_api_call('medidas', params=params)
        if not res: return 0, 0 # ToDo: raise custom error

        canales = res[0]['canales']
        medidas = res[0]['mediciones']
        last_reading = res[0]['fechaUltimaLectura']
        records = [(idPuntoMedida, c['idCanal'], m['intervalo'], m[f'canalVal{c["idCanal"]}']) for c in canales for m in medidas]
        if last_reading: return records, res[0]['fechaUltimaLectura']
        return records

    def get_total_period_energy(self, idPuntoMedida, periodo):
        """
        Get the total energy for every channels for a given measurement point and period.

        Args:
            idPuntoMedida (str): the measure point unique identifier.
            periodo (str): measurement period (monthly granularity) in YYYYMM format.

        Returns: 
            A tuple containing (nan_values, last_reading, net_active_energy, net_reactive_energy). 
            nan_values is a boolean indicating wether there are NaN values in the period or not. 
            last_reading is the last available measurement in YYYY-MM-DD HH:MM:SS format. Usually less
            than 24 hours, but depends on meter connectivity and other variables from the PRMTE service.
            Other two are net energy (injections - withdrawls) for the whole period, as Numpy 
            arrays, in kWh units.
        """
        params = {
            'idPuntoMedida': idPuntoMedida,
            'periodo': periodo,
            'idCanal': '1,2,3,4'
        }
        res = self.make_api_call('medidas', params=params)
        if not res: return 0 # ToDO: Raise custom error 

        medidores = res[0]['medidores']
        last_reading = res[0]['fechaUltimaLectura']
        activa = np.array([medida['canalVal3'] - medida['canalVal1'] if medida['canalVal1'] != None and medida['canalVal3'] != None else np.nan for medida in res[0]['mediciones']])
        reactiva = np.array([medida['canalVal4'] - medida['canalVal2'] if medida['canalVal4'] != None and medida['canalVal2'] != None else np.nan for medida in res[0]['mediciones']])
        nan_values = np.isnan(activa).any() | np.isnan(reactiva).any()
        return nan_values, last_reading, np.nansum(activa), np.nansum(reactiva)


import pandas as pd 
from datetime import datetime, timedelta 
from dateutil.parser import parse as parse_date

def get_missing_dates(df, start_date, end_date, plant_id=None, grouping='minute', granularity=1):
    """
    Gets missing dates in a Pandas DataFrame, checking for a complete index. 
    
    Args:
        df (pd.DataFrame): 
        start_date: 
        end_date: 
        plant_id (int):
        grouping (str):
        granularity (int):

    TO DO:
        - add mapper from 'grouping' and 'granularity' to Pandas frequency. 
    """
    def check_datetime_index(df):
        if isinstance(df.index, pd.DatetimeIndex):
            return True
        elif isinstance(df.index, pd.MultiIndex):
            for level in df.index.levels:
                if isinstance(level, pd.DatetimeIndex):
                    return True
        return False 
                
    if plant_id: 
        df = df.loc[plant_id]

    if isinstance(start_date, str):
        start_date = datetime.strptime(start_date, '%Y-%m-%dT%H:%M:%S')
        end_date = datetime.strptime(end_date, '%Y-%m-%dT%H:%M:%S')

    start_date = floor_and_format(start_date)
    end_date = floor_and_format(end_date)

    if not check_datetime_index(df):
        try:
            df['Date'] = pd.to_datetime(df['Date'])
            df = df.set_index('Date')
        except ValueError:
            raise Exception('DataFrame contains no DatetimeIndex or "Date" column.')

    df = df[(df.index.get_level_values('Date') >= start_date) & (df.index.get_level_values('Date') <= end_date)]

    # Check missing, using default granularity (5MIN) for now
    missing_start, missing_end = check_complete_index(df, start_date, end_date)
    return missing_start, missing_end

def check_complete_index(df, start_date, end_date, freq='5MIN'):
    """
    Checks wether a Pandas DataFrame with DateTime index with name 'Date' is complete or not.
    For now only accepts whole dates with daily granularity! (i.e. 2024-01-25T00:00:00)

    Args:
        start_date (str): date in 'YYYY-MM-DDThh:mm:ss' format
        end_date (str): date in 'YYYY-MM-DDThh:mm:ss' format
        freq (str): date frequency expected in the index as of Pandas documentation (see 
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases)
        
    Returns: 

    """
    # Parse dates 
    start_date = parse_date(start_date)
    end_date = min(parse_date(end_date), datetime.now() - timedelta(minutes=15))

    # Construct expected index
    # By default GPM starts at 00:05:00 for each day and ends at 00:00:00
    expected_range = pd.date_range(start_date + timedelta(minutes=5), end_date, freq=freq)

    # Calculate difference 
    diff = expected_range.difference(df.index.get_level_values('Date'))
    if diff.empty: return None, None

    # Extract only days from missing dates 
    start = diff[0].normalize()
    end = diff[-1].normalize() + timedelta(days=1)

    return start, end

def expected_measurement_count(start_date, end_date, grouping, granularity):
    """ 
    Get expected number of data points depending on the grouping and granularity chosen.
    """
    if isinstance(start_date, str):
        start_datetime = datetime.strptime(start_date, '%Y-%m-%dT%H:%M:%S')
        end_datetime = datetime.strptime(end_date, '%Y-%m-%dT%H:%M:%S')
    else:
        start_datetime = start_date 
        end_datetime = end_date 
    
    delta = end_datetime - start_datetime

    if grouping == 'minute':
        total_minutes = delta.total_seconds() / 60
        return total_minutes / granularity
    elif grouping == 'hour':
        total_hours = delta.total_seconds() / 3600
        return total_hours / granularity
    elif grouping == 'day':
        total_days = delta.days
        return total_days / granularity
    else:
        return 0
    
def split_dates_in_half(start_date, end_date):
    start_datetime = datetime.strptime(start_date, '%Y-%m-%dT%H:%M:%S')
    end_datetime = datetime.strptime(end_date, '%Y-%m-%dT%H:%M:%S')
    midpoint = start_datetime + (end_datetime - start_datetime) / 2
    midpoint_str = midpoint.strftime('%Y-%m-%dT%H:%M:%S')
    return midpoint_str

def floor_and_format(date):
    """ 
    Floors the date's minutes to the closest 5min multiple and formats the date
    to ISO8061 (i.e 2024-01-24T12:15:00)
    """
    round_minutes = (date.minute - date.minute % 5)
    return date.replace(minute=round_minutes, second=0).isoformat(timespec='seconds')


from setuptools import setup, find_packages

setup(
    name='tpower',
    version='0.1.2',
    packages=find_packages(),
    include_package_data=True,
    description='Python package by T-Power.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Matías Galetovic Streeter',
    author_email='matias.galetovic@toesca.com',
    url='https://github.com/toesca-dev/tpower/',
    install_requires=[
        'requests', 'pandas', 'numpy' 'python-dateutil'
    ],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import requests
import logging
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor, as_completed

from data import format_last_data
from dates import expected_measurement_count, split_dates_in_half

class GPMClient:
    BASE_URL = 'https://webapisungrow.horizon.greenpowermonitor.com'

    def __init__(self, username=None, password=None, plant_ids=None, concepts=None, datasources=None):
        self.headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
        self.credentials = {
            'username': username or os.environ.get('GPM_API_USERNAME'),
            'password': password or os.environ.get('GPM_API_PASSWORD')
        }

        if not self.credentials['username'] or not self.credentials['password']:
            raise ValueError("GPM API credentials must be provided or set as environment variables 'GPM_API_USERNAME' and 'GPM_API_PASSWORD'")
        
        self.authenticate()

        self.plant_ids = plant_ids
        self.concepts = concepts
        self.datasources = datasources

    def filter_datasourceids(self, datasourceids=None, plant_id=None, concept=None):
        """
        Used to filter the DataSourceIds to fetch. Can pass either a list or dictionary
        
        """
        data_dict = self.datasources
        if datasourceids is None and plant_id is None and concept is None:
            all_datasources = []
            for plant in data_dict.values():
                for datasources in plant.values():
                    all_datasources.extend(datasources)
            return list(set(all_datasources))
        
        result = []

        if datasourceids:
            if isinstance(datasourceids, list):
                return datasourceids 
            elif isinstance(datasourceids, dict):
                for item in data_dict.values():
                    for key, value in item.items():
                        if key in datasourceids:
                            result.extend(value)
            
        elif isinstance(plant_id, (int, list)) and not concept:
            if isinstance(plant_id, int):
                plant_id = [plant_id]
            for pid in plant_id:
                for item in data_dict.get(pid, {}).values():
                    result.extend(item)
        
        elif isinstance(concept, (str, list)) and not plant_id:
            if isinstance(concept, str):
                concept = [concept]
            for item in data_dict.values():
                for key, value in item.items():
                    if key in concept:
                        result.extend(value)
        
        elif plant_id and concept:
            if isinstance(plant_id, list) and isinstance(concept, list):
                for pid in plant_id:
                    for c in concept:
                        item = data_dict.get(pid, {}).get(c, [])
                        result.extend(item)

            elif isinstance(plant_id, int) and isinstance(concept, str):
                item = data_dict.get(plant_id, {}).get(concept, [])
                result.extend(item)

            elif isinstance(plant_id, int) and isinstance(concept, list):
                for c in concept:
                    item = data_dict.get(plant_id, {}).get(c, [])
                    result.extend(item)

            elif isinstance(plant_id, list) and isinstance(concept, str):
                for pid in plant_id:
                    item = data_dict.get(pid, {}).get(concept, [])
                    result.extend(item)

        return result

    def authenticate(self):
        """
        Internal method for API authentication via bearer token.
        """
        url = self.BASE_URL + '/api/Account/Token'
        response = requests.post(url, headers=self.headers, json=self.credentials)
        if response.status_code == 200:
            res = response.json()
            self.headers['Authorization'] = 'Bearer ' + res['AccessToken']
            logging.info(f'[GPM]: Authentication successful')
        else:
            logging.error(f'[GPM]: Failed to authenticate user. Status code: {response.status_code}')

    def make_api_call(self, endpoint, data):
        """
        Makes a call to the GPM API. Documentation available at 
        https://webapisungrow.horizon.greenpowermonitor.com/swagger/ui/index#/

        Args:
            endpoint (str): back-end endpoint to call
            data (dict): dictionary containing endpoint specific parameters.

        Returns: 
            Response JSON if the call is succesfull, else status code. 
        """
        url = self.BASE_URL + endpoint
        try:
            if endpoint == '/api/Account/Token':
                response = requests.post(url, headers=self.headers, json=data)
            else:
                response = requests.get(url, headers=self.headers, params=data)

            if response.status_code in [200, 206]:
                return response.json()
            elif response.status_code == 416:
                return 416
            else:
                logging.warning(f'[GPM]: Got response with status code {response.status_code}')
                return response.json()
            
        except Exception as e:
            logging.error(f'[GPM]: Exception in API call: {e}')
            return None
        
    def get_plant_information(self):
        """ 
        Get all plants associated to the API user and store PlantIds in a list.
        
        Returns:
            A list of dictionaries, each containing plant specific information.
        """
        res = self.make_api_call('/api/Plant', None)
        self.plant_ids = [plant['Id'] for plant in res]
        return res 
    
    def get_plant_elements(self, plant_id):
        """
        Get all elements associated to a specific PlantId.

        Args:
            plant_id (int): plant identifier (unique for each plant the user has access to)

        Returns:
            A list of dictionaries, each containing an element object.
        """
        res = self.make_api_call(f'/api/Plant/{plant_id}/Element', None)
        return res 
    
    def get_plant_datasources(self, plant_id):
        """
        Get all DataSources associated to a specific PlantId.

        Args:
            plant_id (int): plant identifier (unique for each plant the user has access to)

        Returns:
            A list of dictionaries, each containing a DataSource object.
        """
        res = self.make_api_call(f'/api/Plant/{plant_id}/Datasource', None)
        return res
    
    def get_plant_kpis(self, plant_id):
        """
        Get all KPIs associated to a specific PlantId.

        Args:
            plant_id (int): plant identifier (unique for each plant the user has access to)

        Returns:
            A list of dictionaries, each containing a KPI object.
        """
        res = self.make_api_call(f'/api/Plant/{plant_id}/KPI', None)
        return res
    
    def get_last_data(self, plant_id, datasourceids, tuples=False):
        """
        Gets last available measurements for the given DataSourceIds.

        Args:
            plant_id (int): plant identifier (unique for each plant the user has access to)
            datasourceids (list or dict): DataSourceIds to filter

        Returns:
            If 'datasourceids' is a list, it returns a list of records as per the 'format_last_data' 
            function in the data package. Else, if it's a dictionary containing DataSourceId lists 
            separated per concept, it returns a dictionary of measurements separated per concept.
        """
        res = self.make_api_call(f'/api/Plant/{plant_id}/LastData', None) 

        # TO DO: handle different errors
        if not res: return

        if type(datasourceids) == list:
            formatted_data = format_last_data(res, datasourceids, tuples=tuples)
        elif type(datasourceids) == dict:
            formatted_data = {}
            for concept, ds_ids in datasourceids.items():
                formatted_data[concept] = format_last_data(res, ds_ids, tuples=tuples)
        return formatted_data
    
    def get_last_data_for_all_plants(self):
        """
        Gets last available measurement for all DataSources and all plants. The 'self.datasources' object 
        must be initialized as a dictionary of the form:

            { plant_id: { 'concept': [datasource_ids] } }

        Returns:
            A dictionary containing all measurements per concept, PlantId. It has the form:

                { plant_id: { 'concept': [{'DataSourceId', 'Date', 'Value'}, ...] } }
        
        """
        data = {plant_id: {concept: [] for concept in self.concepts} for plant_id in self.plant_ids}

        with ThreadPoolExecutor() as executor:
            futures = {executor.submit(self.get_last_data, plant_id, self.datasources[plant_id]): plant_id for plant_id in self.plant_ids}

            for future in as_completed(futures):
                plant_id = futures[future]
                try:
                    result = future.result()
                    for key in result:
                        data[plant_id][key].extend(result[key])
                except Exception as exc:
                    print(f'PlantId {plant_id} generated an exception: {exc}')

        return data

    def get_data_list(self, datasources, start_date, end_date, grouping='minute', granularity=1):
        """ 
        Get data list with measurements using the /api/DataList/v2 endpoint. 

        Args:
            datasources (list): List of DataSourceIds to query.
            start_date (str): Start date in ISO8061 format.
            end_date (str): End date in ISO8061 format.
            grouping (str): Time-grouping for response.
            granularity (int): Granularity for the time-grouping parameter.

        Returns:
            The response, which is a list of dictionaries, each representing
            a record of the form:

                {
                    "DataSourceId": 122382,
                    "Date": "2024-01-22T15:00:00.753Z",
                    "Value": 55.43
                }
        """
        params = {
            'startDate': start_date,
            'endDate': end_date,
            'grouping': grouping,
            'granularity': granularity,
            'dataSourceIds': ','.join(map(str, datasources))
        }

        res = self.make_api_call('/api/DataList/v2', params)

        if res == 416: # HTTP 416 - Range not satisfiable
            midpoint = split_dates_in_half(start_date, end_date)
            first_half = self.get_data_list(datasources, start_date, midpoint, grouping, granularity)
            second_half = self.get_data_list(datasources, midpoint, end_date, grouping, granularity)
            return first_half + second_half if first_half and second_half else []
        
        return res

    def get_data_list_in_batches(self, datasources, start_date, end_date, grouping='minute', granularity=5, max_retries=3):
        """ 
        Uses the get_data_list method to fetch data for multiple DataSourceIds. Handles incomplete responses
        (HTTP 206) by making subsequent calls and waiting until all datasources have complete data.

        Args:
            datasources (list): List of DataSourceIds to query.
            start_date (str): Start date in ISO8061 format.
            end_date (str): End date in ISO8061 format.
            grouping (str): Time-grouping for response.
            granularity (int): Granularity for the time-grouping parameter.
            max_retries (int): Max. number of retries in case of 206 response.

        Returns:
            The response, which is a list of dictionaries, each representing
            a record of the form:

                {
                    "DataSourceId": 122382,
                    "Date": "2024-01-22T15:00:00.753Z",
                    "Value": 55.43
                }
        """ 
        def fetch_data(sources):
            """
            Fetches data in parrallel, batching all DataSourceIds into a maximum of 10 per call (API internal limit).
            """
            batches = [sources[i:i + 10] for i in range(0, len(sources), 10)]
            
            # Make parallel calls
            with ThreadPoolExecutor() as executor:
                futures = [executor.submit(self.get_data_list, batch, start_date, end_date, grouping, granularity) for batch in batches]
                results = [future.result() for future in futures]

            # Combine all responses
            combined = []
            for result in results: 
                if result is not None:
                    combined.extend(result)
            return combined

        # Expected number of datapoints for each DataSourceId
        expected_count = expected_measurement_count(start_date, end_date, grouping, granularity)

        retry_count = 0
        complete_data = []
        while datasources and retry_count < max_retries:
            data = fetch_data(datasources)
            complete_data.extend(data)

            # Count datapoints per DataSourceId
            response_counts = {ds: 0 for ds in datasources}
            for item in data:
                if item['DataSourceId'] in response_counts:
                    response_counts[item['DataSourceId']] += 1

            # Check condition and filter already complete DataSourceIds
            datasources = [ds for ds, count in response_counts.items() if count < expected_count]
            retry_count += 1

        return complete_data